{************************************************************}
{                   Delphi VCL Extensions                    }
{                                                            }
{                Drawing complex linetypes                   }
{                                                            }
{      Copyright (c) 2002-2010 SoftGold software company     }
{                                                            }
{************************************************************}

unit sgLines;

{$DEFINE SG_CADIMPORTERDLLDEMO}

interface

uses Dialogs,
  Windows, Graphics, Classes, Math, SysUtils, sgConsts;

type
  PsgLTypeElement = ^TsgLTypeElement;
  TsgLTypeElement = packed record
    // 0 = "Dash" only
    // 1 = If set, Rotation below specifies an absolute rotation (if not set, Rotation specifies a relative rotation- "including on load")
    // 2 = Embedded element is a text string
    // 4 = Embedded element is a shape
    ComplexType: Byte;
    Dash: Double;
    Scale: Double;
    Rotation: Double;
    XOffset: Double;
    YOffset: Double;
    Text: string;
    ShapeNumber: Word;
    case Integer of
    0: (Style: Pointer);
    1: (HStyle: UInt64);
  end;

  TsgDoubleList = class
  private
    FCapacity: Integer;
    FData: PDouble;
    FSize: Integer;
    procedure Grow;
    function GetItem(Index: Integer): Double;
    procedure SetItem(Index: Integer; const Value: Double);
    procedure SetSize(const Value: Integer);
    procedure SetCapacity(const Value: Integer);
  public
    constructor Create(const ASize: Integer);
    constructor CreateEx(const ASize, ACapacity: Integer);
    destructor Destroy; override;
    procedure Add(const Value: Double);
    procedure Assign(ASource: TsgDoubleList);
    procedure Clear;
    procedure Flip;
    property Items[Index: Integer]: Double read GetItem write SetItem; default;
    property Size: Integer read FSize write SetSize;
    property Capacity: Integer read FCapacity write SetCapacity;
  end;

  TsgInt64List = class(TsgDoubleList)
  private
    function GetItem(Index: Integer): Int64;
    function GetObject(Index: Integer): TObject;
    procedure SetItem(Index: Integer; const Value: Int64);
    procedure SetObject(Index: Integer; Value: TObject);
  public
    constructor Create;
    function Add(const Value: Int64): Integer;
    function IndexOf(const Value: Int64): Integer;
    function First: Int64;
    function Last: Int64;
    procedure Insert(Index: Integer; const Value: Int64);
    property Items[Index: Integer]: Int64 read GetItem write SetItem; default;
    property Objects[Index: Integer]: TObject read GetObject write SetObject;
  end;

  TsgLines = class
  private
    FElements: TList;
    FIsUniform: Boolean;
    FPatternLength: Double;
    FScale: Double;
    FSimpleElements: TList;
    procedure AddPt(ADottedPoints: TList; const APoint, ADelta: TFPoint;
      const ACurrentByLength: Double);
    procedure AddPtEx(ADottedPoints: TList; const P1, ADelta, AOffset: TFPoint;
      const ACurrentByLength, AScale, ASin, ACos: Double);
    function CheckPatterns(const ALen: Double; const AAmountPatterns: Integer): Boolean;
    procedure ClearSimpleElements;
    procedure CopyElements(ACopiedElements: TList);
    procedure DoTextPoints(ADottedPoints, APtsListOfList: TList; const P1, ADelta: TFPoint;
      const ACurrentByLength: Double; const AAbsoluteRotation: Boolean);
    function GetAdditionalPart(const ALength, APatternsAmount: Double): Double;
    procedure SetScale(const Value: Double);
    function GetElement(const Index: Integer): PsgLTypeElement;
    function GetElementsCount: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(const AElement: PsgLTypeElement);
    procedure AddTick(const Value: Double);
    procedure Assign(ASource: TsgLines);
    procedure Initialize(const AParts: array of Double);
    procedure Clear;
    procedure Curve(APoints, ADottedPoints: TList; const Close: Boolean);
    procedure Line(const Pt1, Pt2: TFPoint; ADottedPoints: TList);
    procedure Loaded(AConverter: Pointer);
    procedure Poly(APoints, ADottedPoints: TList; const Close: Boolean);
    property Elements[const Index: Integer]: PsgLTypeElement read GetElement;
    property ElementsCount: Integer read GetElementsCount;
    function IsDotted: Boolean;
    function IsSolid: Boolean;
    property IsUniform: Boolean read FIsUniform write FIsUniform;
    property PatternLength: Double read FPatternLength;
    property Scale: Double read FScale write SetScale;
  end;

  function SGFloor(AValue: Double): Integer;

implementation

{$IFNDEF SG_CADIMPORTERDLLDEMO}
uses DXFConv, TTF;

type
  PInt64 = ^Int64;
  TsgDXFConverterAccess = class(TsgDXFConverter);
{$ELSE}
type
{$ENDIF}

  PsgSimpleLTypeElement = ^TsgSimpleLTypeElement;
  TsgSimpleLTypeElement = packed record
    // 0 = "Dash" only
    // 1 = SHX/Shape
    // 2 = TTF
    // 4 = Has absolute rotation
    ComplexType: Byte;
    Points: TList;
    Value: Double;// rotation or dash/dot/space
  end;

{$IFNDEF SG_CADIMPORTERDLLDEMO}
{ GetTextPath
  Creates list of lists of PFPoints text/shape's boundary.

  Returns type of list:
    0 - list not created;
    1 - SHX, Shape;
    2 - TTF.                                                }
function GetTextPath(AConverter: TsgDXFConverter; AText: TsgDXFText;
  ADestList: TList; const AInitialMatrix: PFMatrix): Byte;

  procedure DoTTF;
  var
    vTextGlyph: TsgTextGlyph;
    vMatrix: TFMatrix;
    vIndex: Integer;
  begin
    vIndex := ContainerOfTextGlyphs.IndexOf(AText.Font.Name, TFontStyles(AText.Font.Style));
    if vIndex > -1 then
    begin
      vTextGlyph := ContainerOfTextGlyphs.TextGlyph[vIndex];
      if AInitialMatrix <> nil then
        vMatrix := MatXMat(AInitialMatrix^, AText.GetMatrix)
      else
        vMatrix := AText.GetMatrix;
      vTextGlyph.GetPolyPolylineW(AText.Text, AText.UnicodeText, Byte(AText.Font.Style), @vMatrix, ADestList, nil);
      Result := 2;// TTF
    end;
  end;
begin
  Result := 0;
  if (AConverter <> nil) and AConverter.UseSHXFonts and (not AText.WinFont) and AText.IsSHXFont then
  begin
    AText.GetSHXLinesEx(AConverter.SHXFonts, ADestList, AInitialMatrix, nil, AConverter.CodePage);
    if ADestList.Count > 0 then
      Result := 1   // SHX
    else
      DoTTF;
  end
  else
    DoTTF;
end;
{$ENDIF}

function GetAngleByPoint(const APoint: TFPoint): Double;
var
  vKatetX: Double;
  vCmp: TPoint;
begin
  vCmp := Point(Integer(APoint.X < 0), Integer(APoint.Y < 0));
  vKatetX := Abs(APoint.X);
  if vKatetX < fAccuracy then
    Result := Pi*(0.5 + vCmp.Y)
  else
  begin
    Result := ArcTan(Abs(APoint.Y) / vKatetX);
    case vCmp.X shl 1 + vCmp.Y of
      1:  Result := 2*Pi - Result;
      2:  Result := Pi - Result;
      3:  Result := Pi + Result;
    end;
  end
end;

function SGFloor(AValue: Double): Integer;
begin
  Result := Round(AValue);
  if Result > AValue then
    Result := Result - 1;
end;

{ TsgLines }

constructor TsgLines.Create;
begin
  FElements := TList.Create;
  FSimpleElements := TList.Create;
  FScale := 1;
end;

destructor TsgLines.Destroy;
begin
  Clear;
  ClearSimpleElements;
  FElements.Free;
  FSimpleElements.Free;
  inherited Destroy;
end;

procedure TsgLines.Add(const AElement: PsgLTypeElement);
begin
  FElements.Add(AElement);
end;

{ TsgLines.AddPt

  ACurrentByLength: [0..1], 0 - for starting point, 1 - for ending point.   }
procedure TsgLines.AddPt(ADottedPoints: TList;
  const APoint, ADelta: TFPoint; const ACurrentByLength: Double);
var
  PFPt: PFPoint;
begin
  New(PFPt);
  PFPt^.X := APoint.X + ADelta.X * ACurrentByLength;
  PFPt^.Y := APoint.Y + ADelta.Y * ACurrentByLength;
  PFPt^.Z := APoint.Z + ADelta.Z * ACurrentByLength;
  ADottedPoints.Add(PFPt);
end;

{ TsgLines.AddPtEx

  See TsgLines.AddPt.
  Rotates and offsets a point and adds it to list.    }
procedure TsgLines.AddPtEx(ADottedPoints: TList; const P1, ADelta, AOffset: TFPoint;
  const ACurrentByLength, AScale, ASin, ACos: Double);
var
  PFPt: PFPoint;
begin
  AddPt(ADottedPoints, P1, ADelta, ACurrentByLength);
  PFPt := PFPoint(ADottedPoints.Last);
  PFPt^.X := PFPt^.X + (ACos * AOffset.X - ASin * AOffset.Y)*AScale;
  PFPt^.Y := PFPt^.Y + (ASin * AOffset.X + ACos * AOffset.Y)*AScale;
  PFPt^.Z := PFPt^.Z + (AOffset.Z                          )*AScale;
end;

procedure TsgLines.AddTick(const Value: Double);
var
  vPElement: PsgLTypeElement;
begin
  New(vPElement);
  FillChar(vPElement^, SizeOf(vPElement^), 0);
  vPElement^.Dash := Value;
  Add(vPElement);
end;

procedure TsgLines.Assign(ASource: TsgLines);
begin
  Self.FIsUniform := ASource.FIsUniform;
  Self.FPatternLength := ASource.FPatternLength;
  Self.FScale := ASource.FScale;
  Self.Clear;
  Self.FElements.Capacity := ASource.FElements.Count;
  Self.CopyElements(ASource.FElements);
end;

function TsgLines.IsDotted: Boolean;
begin
  Result := (FSimpleElements.Count > 1) and (PsgSimpleLTypeElement(FSimpleElements.List^[0]).Value = 0)
    and (PsgSimpleLTypeElement(FSimpleElements.List^[1]).Value < 0);
end;

function TsgLines.CheckPatterns(const ALen: Double;
  const AAmountPatterns: Integer): Boolean;
begin
  Result := (ALen = 0) or (AAmountPatterns = 0) or (FPatternLength = 0)
    or (AAmountPatterns > iMaxNumDottedLines) {too many patterns}
    or (FSimpleElements.Count = 0);
end;

procedure TsgLines.Clear;
var
  I: Integer;
begin
  //if FElements = nil then Exit;
  for I := FElements.Count - 1 downto 0 do
    Dispose(PsgLTypeElement(FElements.List^[I]));
  FElements.Clear;
end;

procedure TsgLines.ClearSimpleElements;
var
  I: Integer;
  vPSElem: PsgSimpleLTypeElement;
begin
  //if FSimpleElements = nil then Exit;
  for I := FSimpleElements.Count - 1 downto 0 do
  begin
    vPSElem := PsgSimpleLTypeElement(FSimpleElements.List^[I]);
    FreeRecordListOfList(vPSElem^.Points);
    Dispose(vPSElem);
  end;
  FSimpleElements.Clear;
end;

procedure TsgLines.CopyElements(ACopiedElements: TList);
var
  I, Cnt: Integer;
  vPElem: PsgLTypeElement;
begin
  Cnt := ACopiedElements.Count - 1;
  for I := 0 to Cnt do
  begin
    New(vPElem);
    vPElem^ := PsgLTypeElement(ACopiedElements[I])^;
    FElements.Add(vPElem);
  end;  
end;

procedure TsgLines.DoTextPoints(ADottedPoints, APtsListOfList: TList;
  const P1, ADelta: TFPoint; const ACurrentByLength: Double; const AAbsoluteRotation: Boolean);
var
  S, C: Extended;
  I, J: Integer;
  vList: TList;
begin
  if not AAbsoluteRotation then
    SinCos(GetAngleByPoint(ADelta), S, C)
  else
  begin
    S := 0;
    C := 1;
  end;
  for I := 0 to APtsListOfList.Count - 1 do
  begin
    vList := APtsListOfList.List^[I];
    for J := 0 to vList.Count - 2 do
    begin
      AddPtEx(ADottedPoints, P1, ADelta, PFPoint(vList.List^[J])^, ACurrentByLength, FScale, S, C);
      AddPtEx(ADottedPoints, P1, ADelta, PFPoint(vList.List^[J+1])^, ACurrentByLength, FScale, S, C);
    end;
  end;
end;

function TsgLines.GetAdditionalPart(const ALength, APatternsAmount: Double): Double;
var
  vSElem: TsgSimpleLTypeElement;
begin
  vSElem := PsgSimpleLTypeElement(FSimpleElements[0])^;
  Result := Abs(vSElem.Value) * Ord(vSElem.ComplexType = 0);
  if not FIsUniform then
    Result := 0.5 * (Result + (ALength - APatternsAmount * FPatternLength));
end;

function TsgLines.GetElement(const Index: Integer): PsgLTypeElement;
begin
  if (Index >= 0) and (Index < FElements.Count) then
    Result := PsgLTypeElement(FElements[Index])
  else
    Result := nil;
end;

function TsgLines.GetElementsCount: Integer;
begin
  Result := FElements.Count;
end;

procedure TsgLines.Initialize(const AParts: array of Double);
var
  I: Integer;
  vPElem: PsgLTypeElement;
begin
  Clear;
  for I := Low(AParts) to High(AParts) do
  begin
    New(vPElem);
    FillChar(vPElem^, SizeOf(vPElem^), 0);
    vPElem^.Dash := AParts[I];
    FElements.Add(vPElem);
  end;
end;

function TsgLines.IsSolid: Boolean;
begin
  Result := FElements.Count <= 1;// previous version
end;

procedure TsgLines.Loaded(AConverter: Pointer);
var
{$IFNDEF SG_CADIMPORTERDLLDEMO}
  Converter: TsgDXFConverterAccess absolute AConverter;
  vText: TsgDXFText;
  vList: TList;
  vType: Byte;
{$ENDIF}
  vElem: TsgLTypeElement;
  vPSElem: PsgSimpleLTypeElement;
  I, vElemsCount: Integer;

  procedure NewSElem;
  begin
    New(vPSElem);
    FillChar(vPSElem^, SizeOf(vPSElem^), 0);
  end;
begin
  FPatternLength := 0;
  ClearSimpleElements;
  // Generates new list of simplified records
  vElemsCount := FElements.Count - 1;
  for I := 0 to vElemsCount do
  begin
    vElem := PsgLTypeElement(FElements[I])^;
    //if (I <> 0) and (vElem.ComplexType = 0) and ()  then
    NewSElem;
    vPSElem^.Value := vElem.Dash;
    FPatternLength := FPatternLength + Abs(vElem.Dash);
{$IFNDEF SG_CADIMPORTERDLLDEMO}
    if vElem.ComplexType and 2 <> 0 then
      vText := TsgDXFText.Create
    else
      if (vElem.ComplexType and 4 <> 0) and (vElem.Style <> nil) then
        vText := TsgDXFShape.Create
      else
        vText := nil;
    if vText <> nil then
    begin
      if vElem.Dash <> 0.0 then
        FSimpleElements.Add(vPSElem);
      vText.Point := MakeFPoint(vElem.XOffset, vElem.YOffset, 0);
      vText.Height := vElem.Scale;
      vText.Rotation := vElem.Rotation;
      vText.Text := vElem.Text;
      if vText is TsgDXFShape then
        TsgDXFShape(vText).ShapeNumber := vElem.ShapeNumber;
      vText.Style := vElem.Style;
      if Converter <> nil then
        Converter.Loads(vText);
      // Get polylines here
      // ...
      vList := TList.Create;
      if IsBadRect(vText.Box) then
        vType := 0
      else
        vType := GetTextPath(Converter, vText, vList, nil);
      vText.Free;
      if vType = 0 then // StdMat(MakeFPoint(vElem.Scale, vElem.Scale, 1), cnstFPointZero)
      begin
        vList.Free;
        Continue;
      end
      else
      begin
        NewSElem;
        vPSElem^.ComplexType := vType;
        vPSElem^.Points := vList;
        vPSElem^.Value := vElem.Rotation;
        if vElem.ComplexType and 1 <> 0 then
          vPSElem^.ComplexType := vPSElem^.ComplexType or 4;// Absolute rotation
      end;
    end;
{$ENDIF}
    FSimpleElements.Add(vPSElem);
  end;
  if FPatternLength = 0 then
    ClearSimpleElements;
end;

procedure TsgLines.Poly(APoints, ADottedPoints: TList;
  const Close: Boolean);
var
  I: Integer;
  Pt1, Pt2: TFPoint;

  function EqualPoints(const AP1, AP2: TFPoint): Boolean;
  begin
    Result := (Pt1.X = Pt2.X) and (Pt1.Y = Pt2.Y) and (Pt1.Z = Pt2.Z);
  end;
begin
  for I := 0 to APoints.Count - 2 do
  begin
    Pt1 := PFPoint(APoints.Items[I])^;
    Pt2 := PFPoint(APoints.Items[I+1])^;
    if (I <> 0) and EqualPoints(Pt1, Pt2) then
      Continue;
    Line(Pt1, Pt2, ADottedPoints);
  end;
  if Close and (APoints.Count > 1) then//  Addition of the first point of a polyline
  begin
    Pt1 := PFPoint(APoints.Items[APoints.Count - 1])^;
    Pt2 := PFPoint(APoints.Items[0])^;
    if not EqualPoints(Pt1, Pt2) then
      Line(Pt1, Pt2, ADottedPoints);
  end;
end;

procedure TsgLines.SetScale(const Value: Double);
begin
  FScale := Abs(Value);
  if FScale < fAccuracy then
    FScale := 1.0;
end;

procedure TsgLines.Line(const Pt1, Pt2: TFPoint; ADottedPoints: TList);
var
  AmountPatterns, CounterPatterns, vCount, vIndex: Integer;
  Cur, Len, PartLineLength: Double;
  vSElem: TsgSimpleLTypeElement;
  //DottedPoints: TList;
  Delta: TFPoint;
begin
  //DottedPoints := TList.Create;
  vCount := FSimpleElements.Count-1;
  try
    Delta.X := Pt2.X - Pt1.X;
    Delta.Y := Pt2.Y - Pt1.Y;
    Delta.Z := Pt2.Z - Pt1.Z;
    if FPatternLength <> 0 then
    begin
      Len := Sqrt(Delta.X*Delta.X + Delta.Y*Delta.Y + Delta.Z*Delta.Z) / FScale;
      if Len < iMaxNumDottedLines * FPatternLength then
        AmountPatterns := SGFloor(Len / FPatternLength)
      else
      begin
        Len := 0;
        AmountPatterns := iMaxNumDottedLines + 1;
      end;
    end
    else
    begin
      Len := 0;
      AmountPatterns := 0;
    end;
    if CheckPatterns(Len, AmountPatterns) then
    begin
      AddPt(ADottedPoints, Pt1, Delta, 0);// starting point
      if IsDotted and (AmountPatterns = 0) then
      begin
        AddPt(ADottedPoints, Pt1, Delta, 0);
        AddPt(ADottedPoints, Pt1, Delta, 1);
      end;
      AddPt(ADottedPoints, Pt1, Delta, 1);// ending point
      Exit;
    end;

    Cur := 0;
    vIndex := 0;
    CounterPatterns := 0;

    // Length of a additional part of a line
    PartLineLength := GetAdditionalPart(Len, AmountPatterns);

    while True do
    begin
      vSElem := PsgSimpleLTypeElement(FSimpleElements[vIndex])^;
      case vSElem.ComplexType of
        0:   // Dash/dot/space
          begin
            if vSElem.Value > 0 then
            begin   // dash
              AddPt(ADottedPoints, Pt1, Delta, Cur/Len);
              if (CounterPatterns < AmountPatterns) or FIsUniForm then
                Cur := Cur + vSElem.Value*Ord(Cur <> 0) + PartLineLength*Ord(Cur = 0)
              else
                Cur := Len;
              if Cur > Len then
                Cur := Len;
              AddPt(ADottedPoints, Pt1, Delta, Cur/Len);
            end
            else
              if vSElem.Value = 0 then
              begin // dot
                if CounterPatterns = AmountPatterns then
                  Cur := Len;
                AddPt(ADottedPoints, Pt1, Delta, Cur/Len);
                AddPt(ADottedPoints, Pt1, Delta, Cur/Len);
              end
              else  // space
                Cur := Cur + Abs(vSElem.Value) + PartLineLength*Ord(Cur = 0);
          end;
        1, 5:// SHX, Shape
          begin
            DoTextPoints(ADottedPoints, vSElem.Points, Pt1, Delta, Cur/Len, vSElem.ComplexType and 4 <> 0);
          end;
        2, 6:// TTF
          begin
            DoTextPoints(ADottedPoints, vSElem.Points, Pt1, Delta, Cur/Len, vSElem.ComplexType and 4 <> 0);
          end;
      end;
      if Cur >= Len then Exit;//if Abs(Cur - Len) < fAccuracy then Exit;
      Inc(vIndex);
      if vIndex > vCount then
      begin
        vIndex := 0;
        Inc(CounterPatterns);// another one pattern has been added
      end;
    end;
  finally
    //DottedPoints.Free;
  end;
end;

procedure TsgLines.Curve(APoints, ADottedPoints: TList; const Close: Boolean);
var
  I, vIndex, vPointsCount, vCount, AmountPatterns, CounterPatterns: Integer;
  Cur, TotalCur, Len, LineLength, PartLineLength, vParam, vScale: Double;
  Beginning, isTooManyDottedLines: Boolean;
  vSElem: TsgSimpleLTypeElement;
  Lengths: TsgDoubleList;
  Delta, P1, P2: TFPoint;

  procedure AddPointToList(AList: TList; const APoint: Pointer);
  var
    PFPt: PFPoint;
  begin
    New(PFPt);
    PFPt^ := PFPoint(APoint)^;
    AList.Add(PFPt);
  end;

begin
  vPointsCount := APoints.Count-1;
  if Close and (vPointsCount > 0) then//  Addition of the first point of a curve
  begin
    AddPointToList(APoints, APoints[0]);
    Inc(vPointsCount);
  end;

  isTooManyDottedLines := False;
  Len := 0.0;
  vIndex := 0;
  // Creates the lengths-list of segments of this "polyline"
  Lengths := TsgDoubleList.Create(vPointsCount);
  try
    P1 := PFPoint(APoints[vPointsCount])^;
    for I := vPointsCount-1 downto 0 do
    begin
      P2 := PFPoint(APoints[I])^;
      LineLength := Sqrt((P2.X - P1.X) * (P2.X - P1.X) + (P2.Y - P1.Y) * (P2.Y - P1.Y)+ (P2.Z - P1.Z) * (P2.Z - P1.Z)) {/ FScale};
      if LineLength = 0 then
      begin
        Dispose(APoints[I]);
        APoints.Delete(I);
        Continue;
      end;
      Lengths[vIndex] := LineLength;
      Len := Len + Lengths[vIndex];
      Inc(vIndex);
      P1 := PFPoint(APoints[I])^;
      isTooManyDottedLines := isTooManyDottedLines or (LineLength > iMaxNumDottedLines * FScale * FPatternLength);
    end;

    vPointsCount := APoints.Count-1;

    if vPointsCount < 1 then
    begin
      AddPointToList(ADottedPoints, APoints[0]);
      AddPointToList(ADottedPoints, APoints[0]);
      Exit;
    end;


    if FPatternLength <> 0 then
    begin
      // Calculates amount of patterns in the "polyline"
      AmountPatterns := SGFloor(Len / (FScale*FPatternLength));
      if Close and (AmountPatterns <> 0) then
      begin
        //vScale := FScale * ( 1 + (Len - AmountPatterns * FPatternLength)/AmountPatterns/FPatternLength);
        vScale := Len/(AmountPatterns * FPatternLength);
        Len := AmountPatterns * FPatternLength;
      end
      else
      begin
        vScale := FScale;
        Len := Len/vScale;
      end;
      // Recalculates amount of patterns in the "polyline" (accuracy problem)
      //AmountPatterns := SGFloor(Len / FPatternLength);
      // Length of a additional part of a line
      PartLineLength := GetAdditionalPart(Len, AmountPatterns);
    end
    else
    begin
      vScale := FScale;
      PartLineLength := 0;
      AmountPatterns := 0;
      Len := 0;
    end;

    if CheckPatterns(Len, AmountPatterns) or isTooManyDottedLines
      or (AmountPatterns > iMaxNumDottedLines shl 10) then
    begin
      if IsDotted and (AmountPatterns = 0) and (vPointsCount > 0) then
      begin
        AddPointToList(ADottedPoints, APoints[0]);
        AddPointToList(ADottedPoints, APoints[0]);
        AddPointToList(ADottedPoints, APoints[vPointsCount]);
        AddPointToList(ADottedPoints, APoints[vPointsCount]);
      end
      else
        for I := 0 to vPointsCount - 1 do
        begin
          AddPointToList(ADottedPoints, APoints[I]);
          AddPointToList(ADottedPoints, APoints[I+1]);
        end;
      Exit;
    end;

    I := 0;
    vIndex := 0;
    Beginning := True;
    vCount := FSimpleElements.Count-1;
    P1 := PFPoint(APoints[I])^;
    P2 := PFPoint(APoints[I+1])^;
    LineLength := (Lengths.Items[vPointsCount - 1 - I])/vScale;
    Delta.X := P2.X - P1.X;
    Delta.Y := P2.Y - P1.Y;
    Delta.Z := P2.Z - P1.Z;
    Cur := 0;
    TotalCur := 0;
    CounterPatterns := 0;

    while True do
    begin
      vSElem := PsgSimpleLTypeElement(FSimpleElements[vIndex])^;
      case vSElem.ComplexType of
        0:   // Dash/dot/space
          if vSElem.Value = 0 then // dot
          begin
            if CounterPatterns = AmountPatterns then
            begin
              P1 := PFPoint(APoints[vPointsCount])^;// Last curve point
              Cur := 0;//Cur + PartLineLength;//LineLength;
              TotalCur := Len;
            end;
            AddPt(ADottedPoints, P1, Delta, Cur/LineLength);
            AddPt(ADottedPoints, P1, Delta, Cur/LineLength);
          end
          else                     // dash or space
          begin
            if vSElem.Value > 0 then
            begin
              vParam := vSElem.Value*Ord(not Beginning);
              AddPt(ADottedPoints, P1, Delta, Cur/LineLength);
            end
            else
              vParam := Abs(vSElem.Value);
            Cur := Cur + vParam + PartLineLength*Ord(Beginning);
            // Skip full segments
            while (Cur >= LineLength) or (not FIsUniform and (CounterPatterns = AmountPatterns)) do
            begin
              if vSElem.Value > 0 then
                AddPt(ADottedPoints, P1, Delta, 1);
              TotalCur := TotalCur + LineLength;
              if (TotalCur >= Len) or (I = vPointsCount-1) then Exit;
              if vSElem.Value > 0 then
                AddPt(ADottedPoints, P1, Delta, 1);
              // Proceed to next point
              Cur := Cur - LineLength;
              Inc(I);
              P1 := PFPoint(APoints[I])^;
              P2 := PFPoint(APoints[I+1])^;
              LineLength := (Lengths.Items[vPointsCount - 1 - I])/vScale;
              Delta.X := P2.X - P1.X;
              Delta.Y := P2.Y - P1.Y;
              Delta.Z := P2.Z - P1.Z;
            end;
            if vSElem.Value > 0 then
              AddPt(ADottedPoints, P1, Delta, Cur/LineLength);
            Beginning := False;
          end;
        1, 5:// SHX, Shape
          begin
            DoTextPoints(ADottedPoints, vSElem.Points, P1, Delta, Cur/LineLength, vSElem.ComplexType and 4 <> 0);
          end;
        2, 6:// TTF
          begin
            DoTextPoints(ADottedPoints, vSElem.Points, P1, Delta, Cur/LineLength, vSElem.ComplexType and 4 <> 0);
          end;
      end;
      Inc(vIndex);
      if vIndex > vCount then
      begin
        vIndex := 0;
        Inc(CounterPatterns);// another one pattern has been added
      end;
    end;
  finally
    Lengths.Free;
  end;
end;

{ TsgDoubleList }

procedure TsgDoubleList.Add(const Value: Double);
begin
  //if Result = FCapacity then Grow;
  Items[FSize] := Value;
end;

procedure TsgDoubleList.Assign(ASource: TsgDoubleList);
begin
  SetCapacity(ASource.FCapacity);
  FSize := ASource.FSize;
  CopyMemory(FData, ASource.FData, ASource.FCapacity*SizeOf(Double));
end;

procedure TsgDoubleList.Clear;
begin
  FillChar(FData^, FCapacity*SizeOf(Double), 0);
end;

constructor TsgDoubleList.Create(const ASize: Integer);
begin
  FCapacity := ASize;
  FSize := FCapacity;
  FData := AllocMem(FCapacity*SizeOf(Double));
  Clear;
end;

constructor TsgDoubleList.CreateEx(const ASize, ACapacity: Integer);
begin
  SetCapacity(ACapacity);
  if ASize <= ACapacity then
    FSize := ASize
  else
    FSize := ACapacity;
end;

destructor TsgDoubleList.Destroy;
begin
  FreeMem(FData);
  inherited Destroy;
end;

procedure TsgDoubleList.Flip;
var
  I, J, vCnt: Integer;
  vItem: Double;
begin
  vCnt := FSize shr 1 - 1;
  J := FSize - 1;
  for I := 0 to vCnt do
  begin
    // Swap doubles
    vItem := Items[I];
    Items[I] := Items[J];
    Items[J] := vItem;
    Dec(J);
  end;
end;

function TsgDoubleList.GetItem(Index: Integer): Double;
var
  P: PDouble;
begin
  P := FData;
  Inc(P, Index);
  Result := P^;
end;

procedure TsgDoubleList.Grow;
begin
  SetCapacity(FCapacity + 16);
end;

procedure TsgDoubleList.SetCapacity(const Value: Integer);
begin
  if Value < FSize then
    raise Exception.Create('TsgDoubleArray: Value < FSize');
  if Value <> FCapacity then
  begin
    ReallocMem(FData, Value * SizeOf(Double));
    FCapacity := Value;
  end;
end;

procedure TsgDoubleList.SetItem(Index: Integer; const Value: Double);
var
  P: PDouble;
begin
  if Index = FCapacity then
    Grow;
  if (Index >= FSize) and (Index < FCapacity) then
    FSize := Index+1;
  if (Index >= FCapacity) then
    raise Exception.Create('TsgDoubleArray: (Index > FCapacity)');
  P := FData;
  Inc(P, Index);
  P^ := Value;
end;

procedure TsgDoubleList.SetSize(const Value: Integer);
var
  P: PDouble;
begin
  if Value > FCapacity then
    SetCapacity(Value);
  if Value > FSize then
  begin
    P := FData;
    Inc(P, FSize);
    FillChar(P, (Value - FSize) * SizeOf(Double), 0)
  end
  else
    SetCapacity(Value);
  FSize := Value;
end;

{ TsgInt64List }

constructor TsgInt64List.Create;
begin
  CreateEx(0,16);
end;

function TsgInt64List.Add(const Value: Int64): Integer;
var D: Double absolute Value;
begin
  inherited Add(D);
  Result := Size;
end;

function TsgInt64List.GetItem(Index: Integer): Int64;
var
  D: Double;
  I: Int64 absolute D;
begin
  D := inherited GetItem(Index);
  Result := I;
end;

function TsgInt64List.GetObject(Index: Integer): TObject;
var
  D: Double;
  Obj: TObject absolute D;
begin
  D := inherited GetItem(Index);
  Result := Obj;
end;

function TsgInt64List.IndexOf(const Value: Int64): Integer;
var P: PInt64;
begin
  P := PInt64(FData);
  for Result := 0 to Size - 1 do begin
    if P^ = Value then Exit;
    Inc(P);
  end;
  Result := -1;
end;

function TsgInt64List.First: Int64;
begin
  Result := GetItem(0);
end;

function TsgInt64List.Last: Int64;
begin
  Result := GetItem(Size-1);
end;

procedure TsgInt64List.SetItem(Index: Integer; const Value: Int64);
var D: Double absolute Value;
begin
  inherited SetItem(Index,D);
end;

procedure TsgInt64List.SetObject(Index: Integer; Value: TObject);
var
  D: Double;
  Obj: TObject absolute D;
begin
  D := 0;
  Obj := Value;
  inherited SetItem(Index,D);
end;

procedure TsgInt64List.Insert(Index: Integer; const Value: Int64);
var
  Src,Dst: PDouble;
begin
  if Index > Size then raise EListError.Create('TsgInt64List: Index > Size');
  Add(Value);
  if Index = Size-1 then Exit;
  Src := FData; Inc(Src,Index);
  Dst := Src; Inc(Dst);
  Move(Src^, Dst^, SizeOf(Int64)*(Size-Index-1));
  SetItem(Index,Value);
end;

end.
